//-----------------------------------------------------------------------------
// Copyright (c) 2011 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace XNABillboards
{
    /// <summary>
    /// This Windows XNA application demonstrates how to draw an alpha blended
    /// textured quad that is always facing the camera. This technique is
    /// referred to as billboarding. XNA provides built-in support for
    /// billboarding via the CreateBillboard() and CreateConstrainedBillboard()
    /// methods of the Matrix structure. The difference between the
    /// CreateBillboard() and CreateConstrainedBillboard() methods is that the
    /// former creates a spherical billboard that rotates around a specified
    /// object position and the latter creates a cylindrical billboard that
    /// rotates around a specified axis. In this demo we use billboarding to
    /// draw a clump of grass on a dirt floor. We use the the
    /// CreateConstrainedBillboard() method to orient the billboard to ensure
    /// that the clump of grass appears as though it is planted in the ground.
    /// One limitation of the CreateConstrainedBillboard() method is that when
    /// viewing the billboard from an almost straight down viewpoint the edges
    /// of the textured quad become visible thus ruining the effect.
    /// </summary>
    public class Demo : Microsoft.Xna.Framework.Game
    {
        private static void Main()
        {
            using (Demo demo = new Demo())
            {
                demo.Run();
            }
        }

        private struct Ground
        {
            public const float GROUND_SIZE = 512.0f;

            public Texture2D texture;
            public VertexBuffer vertexBuffer;
            public IndexBuffer indexBuffer;
        }

        private struct Billboard
        {
            public const float BILLBOARD_SIZE = 128.0f;

            public Texture2D texture;
            public VertexBuffer vertexBuffer;
            public IndexBuffer indexBuffer;
        }

        private struct Camera
        {
            public const float DOLLYING_SPEED = 2.0f;
            public const float ROTATION_SPEED = 0.50f;
            public const float MOUSEWHEEL_SPEED = 30.0f;
            public const float TRACKING_SPEED = 0.01f;

            public float offset;
            public Vector2 rotation;
            public Vector2 translate;
            public Vector3 position;
            public Vector3 target;
            public Vector3 viewDir;
            public Quaternion orientation;
            public Matrix viewMatrix;
            public Matrix projectionMatrix;
        }
                        
        private GraphicsDeviceManager graphics;
        private SpriteBatch spriteBatch;
        private SpriteFont spriteFont;
        private KeyboardState currentKeyboardState;
        private KeyboardState prevKeyboardState;
        private MouseState currentMouseState;
        private MouseState prevMouseState;
        private Camera camera;
        private Ground ground;
        private Effect effect;
        private Billboard billboard;
        private int windowWidth;
        private int windowHeight;
        private int frames;
        private int framesPerSecond;
        private TimeSpan elapsedTime = TimeSpan.Zero;
        private TimeSpan mouseIdleElapsedTime = TimeSpan.Zero;
        private bool displayHelp;

        public Demo()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            Window.Title = "XNA 4.0 Billboards Demo";
            IsFixedTimeStep = false;
            IsMouseVisible = true;
        }

        protected override void Initialize()
        {   
            base.Initialize();

            // Setup the window to be a quarter the size of the desktop.
            windowWidth = GraphicsDevice.DisplayMode.Width / 2;
            windowHeight = GraphicsDevice.DisplayMode.Height / 2;

            // Setup frame buffer.
            graphics.SynchronizeWithVerticalRetrace = false;
            graphics.PreferredBackBufferWidth = windowWidth;
            graphics.PreferredBackBufferHeight = windowHeight;
            graphics.PreferMultiSampling = true;
            graphics.ApplyChanges();

            // Setup the initial input states.
            currentKeyboardState = Keyboard.GetState();

            // Create the geometry for the billboard.
            InitBillboard(Billboard.BILLBOARD_SIZE, Billboard.BILLBOARD_SIZE, new Vector3(0.0f, Billboard.BILLBOARD_SIZE * 0.5f, 0.0f));

            // Create the geometry for the ground plane.
            InitGround(Ground.GROUND_SIZE, Ground.GROUND_SIZE);
                        
            // Setup the camera.
            camera.target = Vector3.Zero;
            camera.offset = Ground.GROUND_SIZE * 0.75f;
            camera.orientation = Quaternion.Identity;
            camera.rotation = new Vector2(15.0f, 0.0f);
            UpdateCamera();
        }

        private void InitBillboard(float width, float height, Vector3 position)
        {
            float x = width * 0.5f;
            float y = height * 0.5f;

            Vector3 upperLeft = new Vector3(-x, y, 0.0f);
            Vector3 upperRight = new Vector3(x, y, 0.0f);
            Vector3 lowerLeft = new Vector3(-x, -y, 0.0f);
            Vector3 lowerRight = new Vector3(x, -y, 0.0f);

            VertexPositionTexture[] vertices =
            {
                new VertexPositionTexture(position + upperLeft,  new Vector2(0.0f, 0.0f)),  // 0
                new VertexPositionTexture(position + upperRight, new Vector2(1.0f, 0.0f)),  // 1
                new VertexPositionTexture(position + lowerLeft,  new Vector2(0.0f, 1.0f)),  // 2
                new VertexPositionTexture(position + lowerRight, new Vector2(1.0f, 1.0f)),  // 3
            };

            billboard.vertexBuffer = new VertexBuffer(GraphicsDevice, typeof(VertexPositionTexture), vertices.Length, BufferUsage.WriteOnly);
            billboard.vertexBuffer.SetData(vertices);

            short[] indices =
            {
                0, 1, 2,
                2, 1, 3
            };

            billboard.indexBuffer = new IndexBuffer(GraphicsDevice, IndexElementSize.SixteenBits, indices.Length, BufferUsage.WriteOnly);
            billboard.indexBuffer.SetData(indices);
        }

        private void InitGround(float groundWidth, float groundLength)
        {
            float x = groundWidth * 0.5f;
            float z = groundLength * 0.5f;

            VertexPositionTexture[] vertices =
            {
                new VertexPositionTexture(new Vector3(-x, 0.0f, -z), new Vector2(0.0f, 0.0f)),  // 0
                new VertexPositionTexture(new Vector3( x, 0.0f, -z), new Vector2(1.0f, 0.0f)),  // 1
                new VertexPositionTexture(new Vector3(-x, 0.0f,  z), new Vector2(0.0f, 1.0f)),  // 2
                new VertexPositionTexture(new Vector3( x, 0.0f,  z), new Vector2(1.0f, 1.0f)),  // 3
            };

            ground.vertexBuffer = new VertexBuffer(GraphicsDevice, typeof(VertexPositionTexture), vertices.Length, BufferUsage.WriteOnly);
            ground.vertexBuffer.SetData(vertices);

            short[] indices =
            {
                0, 1, 2,
                2, 1, 3
            };

            ground.indexBuffer = new IndexBuffer(GraphicsDevice, IndexElementSize.SixteenBits, indices.Length, BufferUsage.WriteOnly);
            ground.indexBuffer.SetData(indices);
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);
            spriteFont = Content.Load<SpriteFont>(@"Fonts\DemoFont");

            billboard.texture = Content.Load<Texture2D>(@"Textures\grass");
            ground.texture = Content.Load<Texture2D>(@"Textures\ground");

            effect = Content.Load<Effect>(@"Effects\Albedo");
            effect.CurrentTechnique = effect.Techniques["Albedo"];
        }

        protected override void UnloadContent()
        {
        }

        private void ToggleFullScreen()
        {
            int newWidth = 0;
            int newHeight = 0;

            graphics.IsFullScreen = !graphics.IsFullScreen;

            if (graphics.IsFullScreen)
            {
                newWidth = GraphicsDevice.DisplayMode.Width;
                newHeight = GraphicsDevice.DisplayMode.Height;
            }
            else
            {
                newWidth = windowWidth;
                newHeight = windowHeight;
            }

            graphics.PreferredBackBufferWidth = newWidth;
            graphics.PreferredBackBufferHeight = newHeight;
            graphics.PreferMultiSampling = true;
            graphics.ApplyChanges();
        }

        private bool KeyJustPressed(Keys key)
        {
            return currentKeyboardState.IsKeyDown(key) && prevKeyboardState.IsKeyUp(key);
        }

        private void ProcessKeyboard()
        {
            prevKeyboardState = currentKeyboardState;
            currentKeyboardState = Keyboard.GetState();

            if (KeyJustPressed(Keys.Escape))
                this.Exit();

            if (currentKeyboardState.IsKeyDown(Keys.LeftAlt) || currentKeyboardState.IsKeyDown(Keys.RightAlt))
            {
                if (KeyJustPressed(Keys.Enter))
                    ToggleFullScreen();
            }

            if (KeyJustPressed(Keys.H))
                displayHelp = !displayHelp;
        }

        private void ProcessMouse(GameTime gameTime)
        {
            prevMouseState = currentMouseState;
            currentMouseState = Mouse.GetState();

            camera.translate.X = camera.translate.Y = 0.0f;
            camera.rotation.X = camera.rotation.Y = 0.0f;

            float dx = 0.0f;
            float dy = 0.0f;
            float dz = 0.0f;

            if (currentMouseState.MiddleButton == ButtonState.Pressed)
            {
                dz = currentMouseState.Y - prevMouseState.Y;
                dz *= Camera.DOLLYING_SPEED;

                camera.offset += dz;
            }
            else if (currentMouseState.RightButton == ButtonState.Pressed)
            {
                dx = currentMouseState.X - prevMouseState.X;
                dx *= Camera.ROTATION_SPEED;

                dy = currentMouseState.Y - prevMouseState.Y;
                dy *= Camera.ROTATION_SPEED;

                camera.rotation.X = dy;
                camera.rotation.Y = dx;
            }

            // Process mouse wheel scrolling.

            if (currentMouseState.ScrollWheelValue > prevMouseState.ScrollWheelValue)
                camera.offset -= Camera.MOUSEWHEEL_SPEED;
            else if (currentMouseState.ScrollWheelValue < prevMouseState.ScrollWheelValue)
                camera.offset += Camera.MOUSEWHEEL_SPEED;

            if (camera.offset <= Ground.GROUND_SIZE * 0.1f)
                camera.offset = Ground.GROUND_SIZE * 0.1f;
            else if (camera.offset > Ground.GROUND_SIZE)
                camera.offset = Ground.GROUND_SIZE;

            // Hide the mouse cursor when the mouse hasn't moved.

            int deltaX = currentMouseState.X - prevMouseState.X;
            int deltaY = currentMouseState.Y - prevMouseState.Y;
            bool idle = deltaX == 0 && deltaY == 0;

            if (idle)
            {
                mouseIdleElapsedTime += gameTime.ElapsedGameTime;

                if (mouseIdleElapsedTime > TimeSpan.FromSeconds(2.0))
                    IsMouseVisible = false;
            }
            else
            {
                IsMouseVisible = true;
                mouseIdleElapsedTime = TimeSpan.Zero;
            }
        }

        protected override void Update(GameTime gameTime)
        {
            if (!this.IsActive)
                return;

            base.Update(gameTime);

            ProcessKeyboard();
            ProcessMouse(gameTime);
            UpdateCamera();
            UpdateFrameRate(gameTime);
        }

        private void UpdateCamera()
        {
            // Calculate the view matrix.

            Quaternion rotation = Quaternion.Identity;
            float heading = MathHelper.ToRadians(camera.rotation.Y);
            float pitch = MathHelper.ToRadians(camera.rotation.X);

            if (heading != 0.0f)
            {
                rotation = Quaternion.CreateFromAxisAngle(Vector3.UnitY, heading);
                Quaternion.Concatenate(ref rotation, ref camera.orientation, out camera.orientation);
            }

            if (pitch != 0.0f)
            {
                rotation = Quaternion.CreateFromAxisAngle(Vector3.UnitX, pitch);
                Quaternion.Concatenate(ref camera.orientation, ref rotation, out camera.orientation);
            }

            Matrix.CreateFromQuaternion(ref camera.orientation, out camera.viewMatrix);

            Vector3 xAxis = new Vector3(camera.viewMatrix.M11, camera.viewMatrix.M21, camera.viewMatrix.M31);
            Vector3 yAxis = new Vector3(camera.viewMatrix.M12, camera.viewMatrix.M22, camera.viewMatrix.M32);
            Vector3 zAxis = new Vector3(camera.viewMatrix.M13, camera.viewMatrix.M23, camera.viewMatrix.M33);

            camera.target -= xAxis * camera.translate.X;
            camera.target -= yAxis * camera.translate.Y;

            camera.position = camera.target + zAxis * camera.offset;

            camera.viewMatrix.M41 = -Vector3.Dot(xAxis, camera.position);
            camera.viewMatrix.M42 = -Vector3.Dot(yAxis, camera.position);
            camera.viewMatrix.M43 = -Vector3.Dot(zAxis, camera.position);

            Vector3.Negate(ref zAxis, out camera.viewDir);

            // Calculate the projection matrix.

            camera.projectionMatrix = Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4,
                (float)windowWidth / (float)windowHeight, 0.1f, Ground.GROUND_SIZE * 2.0f);
        }

        private void UpdateFrameRate(GameTime gameTime)
        {
            elapsedTime += gameTime.ElapsedGameTime;

            if (elapsedTime > TimeSpan.FromSeconds(1))
            {
                elapsedTime -= TimeSpan.FromSeconds(1);
                framesPerSecond = frames;
                frames = 0;
            }
        }

        private void IncrementFrameCounter()
        {
            ++frames;
        }

        private void DrawText()
        {
            string text = null;

            if (displayHelp)
            {
                System.Text.StringBuilder buffer = new System.Text.StringBuilder();

                buffer.AppendLine("Right click and drag to orbit camera");
                buffer.AppendLine("Middle click and drag to dolly camera");
                buffer.AppendLine("Mouse wheel to dolly camera");
                buffer.AppendLine();
                buffer.AppendLine("Press ALT + ENTER to toggle full screen");
                buffer.AppendLine("Press H to hide help");

                text = buffer.ToString();
            }
            else
            {
                System.Text.StringBuilder buffer = new System.Text.StringBuilder();

                buffer.AppendFormat("FPS: {0}\n", framesPerSecond);
                buffer.AppendLine();
                buffer.Append("Press H to display help");

                text = buffer.ToString();
            }

            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);
            spriteBatch.DrawString(spriteFont, text, new Vector2(1.0f, 1.0f), Color.Yellow);
            spriteBatch.End();
        }

        private void DrawBillboard()
        {
            GraphicsDevice.SetVertexBuffer(billboard.vertexBuffer);
            GraphicsDevice.Indices = billboard.indexBuffer;

            RasterizerState prevRasterizerState = GraphicsDevice.RasterizerState;
            BlendState prevBlendState = GraphicsDevice.BlendState;

            GraphicsDevice.BlendState = BlendState.NonPremultiplied;
            GraphicsDevice.RasterizerState = RasterizerState.CullNone;

            // Create a matrix that will rotate the billboard so that it will
            // always face the camera.

            Matrix billboardMatrix = Matrix.CreateConstrainedBillboard(Vector3.Zero,
                camera.position, Vector3.Up, camera.viewDir, Vector3.UnitZ);

            effect.Parameters["world"].SetValue(billboardMatrix);
            effect.Parameters["view"].SetValue(camera.viewMatrix);
            effect.Parameters["projection"].SetValue(camera.projectionMatrix);
            effect.Parameters["colorMap"].SetValue(billboard.texture);

            foreach (EffectPass pass in effect.CurrentTechnique.Passes)
            {
                pass.Apply();

                GraphicsDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList,
                    0, 0, billboard.vertexBuffer.VertexCount, 0, 2);
            }

            GraphicsDevice.SetVertexBuffer(null);
            GraphicsDevice.Indices = null;

            GraphicsDevice.BlendState = prevBlendState;
            GraphicsDevice.RasterizerState = prevRasterizerState;
        }
               
        private void DrawGround()
        {
            GraphicsDevice.SetVertexBuffer(ground.vertexBuffer);
            GraphicsDevice.Indices = ground.indexBuffer;

            RasterizerState prevState = GraphicsDevice.RasterizerState;
            
            GraphicsDevice.RasterizerState = RasterizerState.CullNone;
            
            effect.Parameters["world"].SetValue(Matrix.Identity);
            effect.Parameters["view"].SetValue(camera.viewMatrix);
            effect.Parameters["projection"].SetValue(camera.projectionMatrix);
            effect.Parameters["colorMap"].SetValue(ground.texture);

            foreach (EffectPass pass in effect.CurrentTechnique.Passes)
            {
                pass.Apply();
                
                GraphicsDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList,
                    0, 0, ground.vertexBuffer.VertexCount,
                    0, ground.indexBuffer.IndexCount / 3);
            }

            GraphicsDevice.SetVertexBuffer(null);
            GraphicsDevice.Indices = null;

            GraphicsDevice.RasterizerState = prevState;
        }

        protected override void Draw(GameTime gameTime)
        {
            if (!this.IsActive)
                return;

            GraphicsDevice.Clear(Color.CornflowerBlue);

            DrawGround();
            DrawBillboard();
            DrawText();

            base.Draw(gameTime);
            IncrementFrameCounter();
        }
    }
}